import base64
from pathlib import Path

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def get_key(salt: str, password: str) -> bytes:
    # From: https://cryptography.io/en/latest/hazmat/primitives/key-derivation-functions/#nist
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt.encode("utf-8"),
        iterations=100000
    )
    return base64.urlsafe_b64encode(kdf.derive(password.encode("utf-8")))


def verify_file(key: bytes, infile: Path, outfile: Path):
    decrypted = read_decrypted(key, outfile)
    plaintext = infile.open('rb').read()
    return decrypted == plaintext


def read_decrypted(key: bytes, file: Path):
    fernet = Fernet(key)
    ciphertext = file.open('rb').read()
    return fernet.decrypt(ciphertext)


def decrypt_file(key: bytes, infile: Path, outfile: Path):
    """Decrypt infile and save as outfile"""
    plaintext = read_decrypted(key, infile)
    outfile.open('wb').write(plaintext)


def encrypt_file(key: bytes, infile: Path, outfile: Path):
    """Encrypt infile and save as outfile"""
    fernet = Fernet(key)
    plaintext = infile.open('rb').read()
    ciphertext = fernet.encrypt(plaintext)
    outfile.open('wb').write(ciphertext)




import sys

from argparse import Namespace

from compendium.command.command import CompendiumCommand
from compendium.initsegment import SEGMENTS


class Init(CompendiumCommand):
    """
    Setup a new compendium folder
    """

    @classmethod
    def add_arguments(cls, parser):
        for segment_class in SEGMENTS:
            segment_class.add_arguments(parser)

    @classmethod
    def run(cls, args: Namespace):
        for segment_class in SEGMENTS:
            try:
                segment_class.check_arguments(args)
            except Exception as e:
                print(f"Invalid argument: {e}", file=sys.stderr)
                return
        initial_segment = SEGMENTS[0]()  # initial initsegment is responsible for creating compendium
        initial_segment.interactive_arguments(args)
        compendium = initial_segment.run(args)
        for segment_class in SEGMENTS[1:]:
            segment = segment_class(compendium)
            segment.interactive_arguments(args)
            segment.run(args)
        compendium.save()






from argparse import ArgumentParser, Namespace


from compendium.initsegment.segment import Segment

class CheckSegment(Segment):
    """Init segment to run the checks"""
    ARGS = ["skipcheck"]

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument("--skipcheck", action="store_true",
                            help="Skip the check after init")

    def run(self, args: Namespace):
        if not args.skipcheck:
            import compendium.command.check
            compendium.command.check.run_checks(self.compendium)


"""
Dodo Open Science Compendium - Commands to help create and maintain the compendium

Use `compendium init --help` for information on how to create a new compendium
Use `compendium --help` for a list of all available commands
"""
import argparse
import logging
import sys

from compendium.command import init, check, encrypt, COMMANDS, get_command
from compendium.util import AbsolutePath


def main():
    parser = argparse.ArgumentParser(description=__doc__, prog='compendium')
    parser.add_argument("--verbose", "-v", help="Verbose output", action="store_true")
    parser.add_argument("--quiet", "-q", help="Quiet (minimal output)", action="store_true")
    parser.add_argument("--folder", "-f", help="Compendium root folder (default: current folder or parent)",
                        type=AbsolutePath)
    subparsers = parser.add_subparsers(help='Sub commands', dest='command')
    subparsers.required = True
    for command in COMMANDS:
        command.add_subparser(subparsers)

    if len(sys.argv) == 1:
        print(__doc__, file=sys.stderr)
    args = parser.parse_args()
    level = (logging.DEBUG if args.verbose else (logging.WARN if args.quiet else logging.INFO))
    logging.basicConfig(level=level, format='[%(levelname)-5s] %(message)s')

    cmd = get_command(args.command)
    cmd.run(args)


if __name__ == '__main__':
    main()


"""
Init segments (phases)
"""

from compendium.initsegment.checksegment import CheckSegment
from compendium.initsegment.folderstructure import FolderStructureSegment
from compendium.initsegment.github import GithubSegment
from compendium.initsegment.pyenv import PyEnvSegment

SEGMENTS = [
        GithubSegment,
        FolderStructureSegment,
        PyEnvSegment,
        CheckSegment,
    ]



from argparse import Namespace, ArgumentParser
from typing import Sequence, Tuple, Iterable

from compendium.compendium import Compendium


class Segment:
    """
    Class for compendium commands (init, encrypt, etc.).
    I guess this didn't really need to be a class, but this makes it explicit what commands are needed,
    can provide defaults, and allows for easier iteration over all commands
    can provide defaults, and allows for easier iteration over all commands
    """
    ARGS = []

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        """Create any needed arguments"""
        pass

    @classmethod
    def check_arguments(cls, args: Namespace):
        """Check whether all arguments are OK"""
        for arg in cls.ARGS:
            if not hasattr(args, arg):
                setattr(args, arg, None)

    def __init__(self, compendium: Compendium):
        self.compendium = compendium

    def interactive_arguments(self, args: Namespace):
        """Ask for argument values on the command line"""
        pass

    def run(self, args: Namespace):
        """Run this initsegment"""
        pass

    def check(self) -> Iterable[Tuple[str, bool]]:
        return []



import logging
from argparse import ArgumentParser, Namespace
from pathlib import Path
from urllib.request import urlretrieve

from compendium.initsegment.segment import Segment
from compendium.util import yesno


def _download(fn: str, dest: Path):
    dest.parent.mkdir(exist_ok=True)
    url = f"https://raw.githubusercontent.com/vanatteveldt/compendium-dodo/main/templates/{fn}"
    logging.info(f"Downloading {url} to {dest}")
    urlretrieve(url, dest)


class FolderStructureSegment(Segment):
    ARGS=["dodofile", "gitignore", "license", "data"]

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument("--dodofile", nargs="?", const="yes", choices=["yes", "no"],
                            help="Download the template dodo.py file? ")
        parser.add_argument("--gitignore", nargs="?", const="yes", choices=["yes", "no"],
                            help="Download the template .gitignore file?")
        parser.add_argument("--license", choices=["mit", "ccby", "no"],
                            help="Download one of the default license files?")
        parser.add_argument("--create-data", nargs="?", choices=["yes", "private", "encrypted", "no"],
                            help="Create the data folders?", dest="data")
        parser.add_argument("--create-src", nargs="?", choices=["yes",  "no"], const="yes",
                            help="Create source folders?", dest="src")
        parser.add_argument("--example-files", nargs="?", choices=["yes",  "no"], const="yes",
                            help="Download example compendium files?", dest="examplefiles")

    def interactive_arguments(self, args: Namespace):
        data = args.folder / "data"
        if data.exists():
            logging.debug(f"Data folder exists at {data}, skipping data folder creation questions")
        else:
            if not yesno(f"Create the data folder at {data}?", default=True):
                args.data = "no"
            elif not yesno(f"Does the project contain private data (i.e. data that cannot be shared)?", default=False):
                args.data = "yes"
            elif yesno("Can (some of) the private data be shared in encrypted form?", default=True):
                args.data = "encrypted"
            else:
                args.data = "private"
            args.src = "yes" if yesno("Create source folders for analysis scripts?", default=True) else "no"
            if "no" not in (args.src, args.data):
                args.examplefiles = "yes" if yesno("Download example files from template?", default=False) else "no"

        def _askdownload(file: Path):
            return (not file.exists()) and yesno(f"Download the template {file.name} file?", default=True)
        if not args.dodofile:
            args.dodofile = "yes" if _askdownload(args.folder / "dodo.py") else "no"
        if not args.gitignore:
            args.gitignore = "yes" if _askdownload(args.folder / ".gitignore") else "no"
        if not args.license:
            if (args.folder / "LICENSE").exists():
                logging.debug("LICENSE file exists, skipping license selection")
            elif yesno("Download a template license file to let others know whether they can reuse your material?",
                       default=True):
                print("We have templates for the MIT and CC-BY licenses.")
                print("For more information, see https://opensource.org/licenses")
                if yesno("Use the MIT license? (great for sharing code with minimal strings attached)",
                         default=True):
                    args.license = "mit"
                elif yesno("Use the CC-BY license? (attribution is required, best for reports and documentation)",
                           default=True):
                    args.license = "ccby"
                else:
                    print("No license selected. Please go to https://opensource.org/licenses ")

    def run(self, args: Namespace):
        data = args.folder / "data"
        if args.data != "no":
            folders = ["", "raw", "intermediate"]
            if args.data in {"private", "encrypted"}:
                folders += ["raw-private"]
            if args.data == "encrypted":
                folders += ["raw-private-encrypted"]
            folders = [f for f in folders if not (data/f).exists()]
            if folders:
                logging.info(f"Creating data folders {data}/{folders}")
                for folder in folders:
                    (data / folder).mkdir()
        if args.src == "yes":
            src = args.folder / "src"
            folders = ["", "data-processing", "analysis"]
            logging.info(f"Creating source folders {src}/{folders}")
            for folder in folders:
                (src / folder).mkdir()
        if args.examplefiles == "yes":
            _download("example.py", self.compendium.folders.SRC_PROCESSING / "example.py")
            _download("example2.py", self.compendium.folders.SRC_PROCESSING / "example2.py")
            _download("secret.txt", self.compendium.folders.DATA_PRIVATE / "secret.txt")
        if args.dodofile == "yes":
            _download("dodo.py", args.folder / "dodo.py")
        if args.gitignore == "yes":
            _download("gitignore", args.folder / ".gitignore")
        licensefile = args.folder / "LICENSE"
        if not licensefile.exists():
            if args.license == "mit":
                _download("LICENSE-MIT", licensefile)
            if args.license == "ccby":
                _download("LICENSE-CC-BY", licensefile)

    def check(self):
        yield "Data Folders", (self.compendium.root / "data").exists()
        yield "License file", (self.compendium.root / "LICENSE").exists()
        yield "dodo.py file", (self.compendium.root / "dodo.py").exists()






def document_readme() -> str:
    actions = list(get_actions())
    md = "# Data processing scripts"
    md += "\n\nThis folder contains the following scripts:\n\n"
    for action in actions:
        inputs =  ",".join(f"[{f.name}]({f})" for f in action.inputs)
        targets = ",".join(f"[{f.name}]({f})" for f in action.targets)
        md += f"- [{action.file.name}](action.file): [{inputs} -> {targets}]  \n  {action.headers.get('DESCRIPTION')}  \n  \n"
    return md


def document_process() -> bytes:
    actions = list(get_actions())
    nodes, nodemap, edges = [], {}, []

    def get_node(file):
        file = file.absolute()
        if contained_in(DATA_ENCRYPTED, file):
            shape = "box3d"
            file = file.relative_to(DATA)
        if contained_in(DATA, file):
            shape = "note"
            file = file.relative_to(DATA)
        elif contained_in(SRC_PROCESSING, file):
            shape = "cds"
            file = file.relative_to(SRC_PROCESSING)
        elif contained_in(SRC_ANALYSIS, file):
            shape = "component"
            file = file.relative_to(SRC_ANALYSIS)

        if file not in nodemap:
            name = f"n_{len(nodemap)}"
            nodemap[file] = name
            label = str(file).replace("/", "/\\n")
            nodes.append(f'\n{name} [label="{label}", shape="{shape}"];')
        return nodemap[file]

    for inf in get_files(DATA_ENCRYPTED, ".gpg"):
        outf = DATA_PRIVATE/inf.stem
        node = get_node(inf)
        node2 = get_node(outf)
        edges.append(f'\n{node} -> {node2};')

    for i, action in enumerate(actions):
        node = get_node(action.file)
        for f in action.inputs:
            node2 = get_node(f)
            edges.append(f'\n{node2} -> {node};')
        for f in action.targets:
            node2 = get_node(f)
            edges.append(f'\n{node} -> {node2};')
    nodes = "\n".join(nodes)
    edges = "\n".join(edges)
    dot = f'digraph G {{graph [rankdir="LR"]; \n{nodes}\n\n{edges}\n}}\n'
    return pipe(["dot", "-T", "png"], dot.encode("utf-8"))


def do_document(args):
    filename = args.filename or {"readme": "README.md", "process": "process.png"}[args.what]
    file = Path.cwd()/filename
    if file.exists() and not args.overwrite:
        answer = input(f"File {file} exists, overwrite? [y/N] ")
        answer = answer.lower()[:1]
        if answer == "n":
            return
        elif answer not in ("", "y"):
            print("Could not understand answer, sorry")
            return
    if args.what == "readme":
        text = document_readme()
        with file.open(mode="w") as f:
            f.write(text)
    elif args.what == "process":
        bytes = document_process()
        with file.open(mode="wb") as f:
            f.write(bytes)


def pipe(command, input: bytes, **kargs) -> bytes:
    proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, **kargs)
    out, _err = proc.communicate(input)
    return out


import logging
from argparse import Namespace
from configparser import ConfigParser, NoSectionError, NoOptionError
import crypt
from pathlib import Path
from typing import Optional, Iterable

from cryptography.fernet import InvalidToken
from doit.exceptions import TaskFailed

from compendium.action import Action
from compendium.encryption import get_key, decrypt_file
from compendium.util import get_files, get_headers, parse_files, call

CONFIGFILE = ".compendium.cfg"

EXT_SCRIPT = {".py", ".R", ".Rmd", ".sh"}

class Folders:
    def __init__(self, root: Path):
        self.ROOT = root
        self.DATA = data = root/"data"
        self.DATA_PRIVATE = data/"raw-private"
        self.DATA_ENCRYPTED = data/"raw-private-encrypted"
        self.DATA_RAW = data/"raw"
        self.DATA_INTERMEDIATE= data/"intermediate"

        self.SRC = src = root/"src"
        self.SRC_PROCESSING = src/"data-processing"
        self.SRC_ANALYSIS = src/"analysis"


def find_root(folder: Path) -> Path:
    if folder is None:
        raise ValueError("Compendium folder cannot be None")
    for f in [folder] + list(folder.parents):
        if (f / CONFIGFILE).exists():
            return f
    raise FileNotFoundError(f"Cannot find compendium folder (starting from {folder})")


class Compendium:
    def __init__(self, folder: Path = None, create_new_config=False):
        if folder is None:
            folder = Path.cwd()
        self.cf = ConfigParser()
        if create_new_config:
            self.changed = True
            self.root = folder
        else:
            self.changed = False
            self.root = find_root(folder)
            logging.info(f"Reading configuration file {self.root / CONFIGFILE}")
            self.cf.read(self.root / CONFIGFILE)
        self.folders = Folders(self.root)

    # **** Configuration file management ****

    def save(self, force=False):
        if not (force or self.changed):
            return
        config = self.root / CONFIGFILE
        logging.info(f"Writing compendium configuraton to {config}")
        with config.open('w') as f:
            f.write("# Compendium configuration file\n")
            self.cf.write(f)
        self.changed = False

    def set(self, section: str, option: str, value: str):
        if not self.cf.has_section(section):
            self.cf.add_section(section)
        val = self.get(section, option)
        if val != value:
            logging.debug(f"SET [{section}] {option} to {value} (was:{val})")
            self.cf.set(section, option, value)
            self.changed = True

    def get(self, section: str, option: str, default: str = None) -> str:
        return self.cf.get(section, option, fallback=default)

    @property
    def salt(self) -> str:
        salt = self.get(section="encryption", option="salt")
        if not salt:
            salt = crypt.mksalt()
            self.set("encryption", "salt", salt)
        return salt

    @property
    def pyenv(self) -> Path:
        env = self.cf.get("python", "env", fallback=None)
        if env:
            return self.root / env

    @pyenv.setter
    def pyenv(self, env: Path):
        self.set("python", "env", str(env.relative_to(self.root)))

    def _section(self, section: str):
        if not self.cf.has_section(section):
            self.cf.add_section(section)
        return self.cf[section]

    # **** Tasks and actions ****

    def get_actions(self) -> Iterable[Action]:
        """Yield all processing and analysis scripts"""
        files = (get_files(self.folders.SRC_PROCESSING, suffix=EXT_SCRIPT) +
                 get_files(self.folders.SRC_ANALYSIS, suffix=EXT_SCRIPT))
        for file in files:
            headers = dict(get_headers(file))
            if "CREATES" in headers and "COMMAND" in headers:
                targets = parse_files(headers["CREATES"])
                inputs = parse_files(headers.get("DEPENDS"))
                # build action
                action = f"{headers['COMMAND']} {file}"
                if headers.get("PIPE", "F")[0].lower() == "t":
                    if len(inputs) > 1 or len(targets) > 1:
                        raise ValueError("File {file}: Cannot use PIPE with multiple inputs or outputs")
                    if inputs:
                        action = f"{action} < {inputs[0]}"
                    action = f"{action} > {targets[0]}"
                if file.suffix == ".py" and self.pyenv:
                    # Activate virtual environent before calling script
                    action = f"(. {self.pyenv}/bin/activate; {action})"
                action = f'{action} && echo "[OK] {file.name} completed" 1>&2'
                yield Action(file, action, targets, inputs, headers)

    def decrypt_file_task(self, password: str, source: Path, target: Path):
        if password is None:
            return TaskFailed("No passphrase specified; please use doit passphrase=**** decrypt")
        target.parent.mkdir(exist_ok=True)
        key = get_key(self.salt, password)
        try:
            decrypt_file(key, source, target)
        except InvalidToken:
            return TaskFailed("Incorrect password, could not decrypt files")

    def install_python_task(self):
        from compendium.initsegment.pyenv import install_pyvenv
        install_pyvenv(self.pyenv)




import logging
import sys

from argparse import Namespace
from collections import defaultdict

from compendium.command.command import CompendiumCommand
from compendium.compendium import Compendium
from compendium.initsegment import SEGMENTS
from compendium.util import contained_in

_CHECK_OK = '\u2714\u2009'
_CHECK_FAIL = '\u2718\u2009'


class Check(CompendiumCommand):
    """
    Generic checks for compendium completeness and consistence
    """

    @classmethod
    def run(self, args: Namespace):
        compendium = Compendium(args.folder)
        run_checks(compendium)


def run_checks(compendium: Compendium):
    for segment in SEGMENTS:
        for check, outcome in segment(compendium).check():
            print(f"[{_CHECK_OK if outcome else _CHECK_FAIL}] {check}")


def get_cycles(graph):
    def cycles_node(graph, node, visited=None):
        # Can I find a cycle in the depth-first graph starting from this node?
        if visited is None:
            visited = set()
        for neighbour in graph.get(node, []):
            #print(f"{node} -> {neighbour} (visited: {visited})")
            if neighbour in visited:
                yield neighbour
            else:
                visited.add(neighbour)
                yield from cycles_node(graph, neighbour, visited)

    # for any node, can you find a cycle?
    for n in graph:
        yield from cycles_node(graph, n)


def do_check(args):
    """
    Run sanity checks on the package
    """
    logging.info("Checking consistency of dependency graph")
    inputs, outputs, graph = set(), set(), defaultdict(set)
    for action in get_actions():
        inputs |= set(action.inputs)
        outputs |= set(action.targets)
        for output in action.targets:
            for input in action.inputs:
                graph[input].add(output)
    errors = []
    # check: all inputs need to be either in raw and exist, in private_raw, or in outputs
    for input in inputs - outputs:
        if contained_in(DATA_RAW, input):
            errors.append(f"Input file {input} does not exist")
        elif not contained_in(DATA_PRIVATE, input):
            errors.append(f"Intermediate file {input} is not produced by any script")
    # check that graph does not contain any cycles
    cycles = set(get_cycles(graph))
    for cycle in cycles:
        errors.append(f"Cyclical dependency for file {cycle}")
    if errors:
        print("Package checking resulted in one or more errors:", file=sys.stderr)
        for error in errors:
            print(f"- {error}", file=sys.stderr)
        sys.exit(1)


import logging
import subprocess
from argparse import ArgumentParser, Namespace

#from compendium.command.init import find_env
from pathlib import Path

from compendium.initsegment.segment import Segment
from compendium.util import AbsolutePath, yesno, call


class PyEnvSegment(Segment):
    ARGS = ["python_env"]

    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument("--python-env", nargs="?", type=AbsolutePath, const='env',
                            help="Create a python virtual environment (specify name or usef default 'env')")

    def run(self, args: Namespace):
        if not args.python_env:
            return
        self.compendium.pyenv = args.python_env
        if not args.python_env.exists():
            install_pyvenv(args.python_env)

    def interactive_arguments(self, args: Namespace):
        if not args.python_env:
            if self.compendium.pyenv:
                args.python_env = self.compendium.pyenv
            else:
                env = find_env(self.compendium.root)
                if env and yesno(f"Use the virtual environment at {env}?", default=True):
                    args.python_env = env  # use the python env
                elif yesno("Create a python virtual environment?", default=True):
                    env = input("Name of the folder (leave empty to use 'env'): ").strip()
                    args.python_env = self.compendium.root / (env or "env")


def find_env(location: Path):
    for name in "env", "venv":
        path = location/name
        if (path/"pyvenv.cfg").exists():
            return path


def install_pyvenv(env: Path):
    logging.info(f"Creating Python virtual environment at {env}")
    call(f"python3 -m venv {env}")
    logging.debug("Installing compendium-dodo to virtual environment")
    call(f"{env}/bin/pip install -U pip wheel")
    call(f"{env}/bin/pip install compendium-dodo")
    setup = env.parent/"setup.py"
    if setup.exists():
        call(f"{env}/bin/pip install -e {setup.parent}")
    req = env.parent/"requirements.txt"
    if req.exists():
        call(f"{env}/bin/pip install -r {req}")



from argparse import Namespace
from pathlib import Path

from compendium.compendium import Compendium


class CompendiumCommand:
    """
    (static) Class for compendium commands (init, encrypt, etc.).
    I guess this didn't really need to be a class, but this makes it explicit what commands are needed,
    can provide defaults, and allows for easier iteration over all commands
    Note that by default, this docstring will be displayed as help text for the command.
    """
    name: str = None

    @classmethod
    def get_name(cls):
        return cls.name if cls.name is not None else cls.__name__.lower()

    @classmethod
    def add_subparser(cls, subparsers):
        parser = subparsers.add_parser(cls.get_name(), help=cls.__doc__)
        cls.add_arguments(parser)

    @classmethod
    def add_arguments(cls, parser):
        """Add any extra arguments"""
        pass

    @classmethod
    def run(cls, args: Namespace):
        compendium = Compendium(args.folder or Path.cwd())
        cls.do_run(compendium, args)
        compendium.save()

    @classmethod
    def do_run(cls, compendium: Compendium, args: Namespace):
        pass


"""
Encrypt files in the raw-private folder to raw-private-encrypted
"""

import logging
import sys
from argparse import Namespace
from getpass import getpass
from pathlib import Path
from typing import Optional, Iterable


from compendium.command.command import CompendiumCommand
from compendium.compendium import Compendium
from compendium.encryption import encrypt_file, verify_file, get_key


class Encrypt(CompendiumCommand):
    """Encrypt files from private-raw to private-raw-encrypted"""

    @classmethod
    def add_arguments(cls, parser):
        parser.add_argument("files", nargs="*",
                            help="Specify the files to be encryped (default=all private files)")
        parser.add_argument("--password",
                            help="Specify the password")
        parser.add_argument("--verify", action="store_true",
                            help="Test whether all files are correctly encrypted with this password")

    @classmethod
    def do_run(cls, compendium: Compendium, args: Namespace):
        def _l(f: Path) -> Path:
            return f.relative_to(compendium.root)
        if args.verify:
            for file in compendium.folders.DATA_ENCRYPTED.glob("*"):
                infile = compendium.folders.DATA_PRIVATE/file.name
                if not infile.exists():
                    logging.warning(f"WARNING: Encrypted file {_l(file)} "
                                    f"has no corresponding file in {_l(compendium.PRIVATE)}")

        files = list(get_files(compendium.folders.DATA_PRIVATE, args.files))
        if not files:
            print("No files to encrypt, exiting", file=sys.stderr)
            sys.exit(1)
        if not args.password:
            args.password = getpass("Please specify the password to use: ").strip()
        if not args.password:
            print("No password given, aborting", file=sys.stderr)
            sys.exit(1)
        if not compendium.folders.DATA_ENCRYPTED.exists():
            logging.debug(f"Creating {compendium.folders.DATA_ENCRYPTED}")
            compendium.folders.DATA_ENCRYPTED.mkdir()
        key = get_key(compendium.salt, args.password)
        action = 'Encrypting' if not args.verify else 'Verifying'
        logging.info(f"{action} {len(files)} file(s) from {_l(compendium.folders.DATA_PRIVATE)}")
        for file in files:
            outfile = compendium.folders.DATA_ENCRYPTED/file.name
            if args.verify:
                logging.debug(f".. {_l(outfile)} -> {_l(file)}?")
                if not outfile.exists():
                    print(f"WARNING: File {_l(outfile)} does not exist")
                elif not verify_file(key, file, outfile):
                    print(f"WARNING: File {_l(file)} could not be decrypted from {_l(outfile)}", file=sys.stderr)
            else:
                logging.debug(f".. {file} -> {outfile}")
                encrypt_file(key, file, outfile)


def get_files(folder: Path, files: Optional[Iterable[str]]):
    if not files:
        yield from folder.glob("*")
    else:
        for file in files:
            if file.startswith("/"):  # absolute path
                file = Path(file)
            elif "/" in file: # relative wrt current folder
                file = Path.cwd()/file
            else:  # relative wrt private data folder
                file = folder/file
            if file.parent != folder:
                raise Exception(f"Cannot encrypt {file}, not in {folder}")
            yield file


import subprocess
import sys
import re
from argparse import ArgumentParser, Namespace
from pathlib import Path

import requests
import logging

from compendium.compendium import Compendium, find_root, CONFIGFILE
from compendium.initsegment.segment import Segment
from compendium.util import yesno, AbsolutePath


class GithubSegment(Segment):
    @classmethod
    def add_arguments(cls, parser: ArgumentParser):
        parser.add_argument("--github",
                        help="Link compendium to github repository (URL or username/repository)")

    @classmethod
    def check_arguments(cls, args: Namespace):
        super().check_arguments(args)
        for arg in 'github', 'folder':
            if not hasattr(args, arg):
                setattr(args, arg, None)
        if not hasattr(args, 'folder'):
            args.folder = None
        if args.github and not check_github(args.github):
            raise ValueError(f"Github repository invalid: {args.github}")

    def __init__(self, compendium: Compendium = None):
        """Allow compendium to be None for this segment"""
        super().__init__(compendium)

    def interactive_arguments(self, args: Namespace):
        # Can we find the root folder?
        folder = args.folder or Path.cwd()
        try:
            args.folder = find_root(folder)
            logging.info(f"Found project root folder at {args.folder}")
        except FileNotFoundError:
            if (folder/".git").exists():
                if yesno(f"{folder} is a git repository, use it as the project folder?",
                         default=True):
                    args.github = None  # No need to clone/initialize github repo
                    args.folder = folder
                    return
        if args.folder and (args.folder / ".git").exists():
            if args.github:
                logging.info(f"Compendium folder {args.folder} is already git repository, so ignoring github link")
                args.github = None
            return

        # Check whether user wants to use github
        if (not args.github) and yesno("Link the compendium to a github repository?", default=True):
            print("Note: If you haven't created a repository yet, you can do so now at https://github.com/new")
            args.github = get_github_name()
        if args.folder and args.github:
            if args.folder.name != github_folder_name(args.github):
                if not yesno(f"Folder name {args.folder.name} does not match "
                             f"github repository name {github_folder_name(args.github)}, "
                             f"are you sure you want to proceed", default=True):
                    print("Aborted", file=sys.stderr)
                    sys.exit(1)

        # Determine folder name if needed
        if not args.folder:
            if args.github:
                repo = Path.cwd()/github_folder_name(args.github)

                question = (f"Use existing folder {repo}?" if repo.exists()
                            else f"Clone github repository into new folder {repo}?")
                if yesno(question, default=True):
                    args.folder = repo
                if (repo/".git").exists():
                    logging.info(f"Folder {repo} is already a git repository, so ignoring github link")
                    args.github = None
        if not args.folder:
            if yesno(f"Use current folder {Path.cwd()}?", default=True):
                args.folder = Path.cwd()
            else:
                name = input("Name of the folder to use (will be created if needed, omit to cancel): ")
                if name:
                    args.folder = Path.cwd()/name
                else:
                    print("Aborted", file=sys.stderr)
                    sys.exit(1)

    def run(self, args: Namespace) -> Compendium:
        if args.github:
            if not args.github.startswith("https://"):
                args.github = f"https://github.com/{args.github}"
            if args.folder.exists():
                if (args.folder/".git").exists():
                    logging.warning(f"Folder {args.folder} is already a git repository, ignoring github link")
                logging.info(f"Linking {args.folder} to github repository {args.github}")
                subprocess.check_call("git init", shell=True, cwd=args.folder)
                subprocess.check_call(f"git remote add origin {args.github}", shell=True, cwd=args.folder)
                subprocess.check_call("git fetch", shell=True, cwd=args.folder)
                subprocess.check_call("git checkout --track origin/main", shell=True, cwd=args.folder)
            else:
                logging.info(f"Cloning {args.github} to new folder {args.folder}")
                subprocess.check_call(f"git clone {args.github} {args.folder}", shell=True)

        else:
            if args.folder.exists():
                logging.info(f"Using compendium folder at {args.folder}")
            else:
                logging.info(f"Creating compendium folder at {args.folder}")
                args.folder.mkdir()

        if (args.folder/CONFIGFILE).exists():
            self.compendium = Compendium(args.folder)
        else:
            self.compendium = Compendium(args.folder, create_new_config=True)
        return self.compendium



def check_github(repository: str) -> str:
    """Check if the repository is valid and exists, returning full URL"""
    if not repository.startswith("https://"):
        if not re.match(r"\w+/\w+", repository):
            raise ValueError("Invalid repository name, format should be full repository URL or userame/repository")
        repository = f"https://github.com/{repository}"
    resp = requests.head(repository)
    if resp.status_code != 200:
        raise ValueError(
            f"Repository does not exist or access denied (url: {repository}, status: {resp.status_code}")
    return repository


def github_folder_name(repository: str) -> str:
    name = repository.rsplit("/", maxsplit=1)[-1]
    if name.endswith(".git"):
        name = name[:-len(".git")]
    return name


def get_github_name():
    while True:
        name = input(
            "Name of an existing github repository to link to (url or username/repository, leave empty to cancel): ").strip()
        if not name:
            return
        try:
            return check_github(name)
        except ValueError as e:
            print(str(e), file=sys.stderr)
            continue


from pathlib import Path
from typing import NamedTuple, List, Dict


class Action(NamedTuple):
    file: Path
    action: str
    targets: List[Path]
    inputs: List[Path]
    headers: Dict[str, str]



#! /usr/bin/env python3

"""
CCS Compendium - Commands to help create and maintain an open science compendium
"""

from __future__ import print_function
import sys

from setuptools import setup

setup(name='ccs-compendium',
      description='CCS Compendium - creating a compendium for reproducible science',
      version='0.0.7',
      license='MIT',
      author='Wouter van Atteveldt',
      author_email='wouter@vanatteveldt.com',
      url='http://github.com/ccs-amsterdam/compendium',
      classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        ],
      packages=['compendium', 'compendium.command', 'compendium.initsegment'],
      python_requires='>=3.6',
      install_requires=["doit", "cryptography", "requests"],
      long_description=__doc__,
      entry_points={
          'console_scripts': [
              'compendium = compendium.__main__:main'
          ]
      },
      )



from compendium.command.check import Check
from compendium.command.encrypt import Encrypt
from compendium.command.init import Init

COMMANDS = [
    Init,
    Check,
    Encrypt,
]


def get_command(name: str):
    for cmd in COMMANDS:
        if cmd.get_name() == name:
            return cmd
    raise ValueError(f"Unknown command: {name}")



import logging
import re
import subprocess
from pathlib import Path
from typing import List, Iterable, Tuple


def get_files(folder: Path, suffix=None) -> List[Path]:
    """Get all files contained in Path (optionally filtering for suffix)"""
    if isinstance(suffix, str):
        suffix = [suffix]
    path = Path.cwd() / folder
    if not path.is_dir():
        logging.warning(f"{path} does not exist or is not a directory")
        return []
    return [f for f in path.iterdir() if (f.is_file() and ((suffix is None) or (f.suffix in suffix)))]


def contained_in(parent: Path, descendant: Path) -> bool:
    """Is the second path actually a descendant of the first?"""
    if parent.is_absolute():
        descendant = descendant.absolute()
    return str(descendant.absolute()).startswith(f"{parent}{os.path.sep}")


def get_headers(file: Path) -> Iterable[Tuple[str, str]]:
    """Get [#! command] and [#key: value] headers from a file"""
    for i, line in enumerate(file.open()):
        if not line.strip():
            continue
        if not line.startswith("#"):
            break
        if i == 0 and line.startswith("#!"):
            yield "COMMAND", line[2:].strip()
        m = re.match(r"#(\w+?):(.*)", line)
        if m:
            yield m.groups()[0].strip(), m.groups()[1].strip()


def parse_files(text: str) -> List[Path]:
    """Convert comma-delimited filenames to a list of Path objects"""
    if not text or not text.strip():
        return []
    return [Path(x.strip()) for x in re.split("[ ,]+", text)]


def AbsolutePath(*args, **kargs) -> Path:
    """Create an absolute path (useful as argument 'type')"""
    return Path(*args, **kargs).absolute()


def yesno(prompt, default: bool = None, add_options=True) -> bool:
    """Input prompt, default, and validation for Y/N questions"""
    if add_options:
        if default is None:
            extra = "[y/n]"
        elif default:
            extra = "[Y/n]"
        else:
            extra = "[y/N]"
        prompt = f"{prompt} {extra}"

    result = input(prompt).lower().strip()
    if (not result) and (default is not None):
        return default
    if result.startswith("y"):
        return True
    if result.startswith("n"):
        return False
    return yesno(prompt, default, add_options=False)


def call(cmd: str):
    """Print and call a system command"""
    logging.debug(cmd)
    subprocess.check_call(cmd, shell=True)